package com.gxzc.zen.orm.aop

import com.gxzc.zen.common.util.SpringContextHolder
import com.gxzc.zen.orm.DynamicMultipleDataSource
import com.gxzc.zen.orm.annotation.DynamicDataSource
import com.gxzc.zen.orm.annotation.MultiTransactional
import com.gxzc.zen.orm.contants.DSKey
import org.aspectj.lang.JoinPoint
import org.aspectj.lang.ProceedingJoinPoint
import org.aspectj.lang.annotation.*
import org.aspectj.lang.reflect.MethodSignature
import org.slf4j.LoggerFactory
import org.springframework.core.annotation.Order
import org.springframework.stereotype.Component
import org.springframework.transaction.jta.JtaTransactionManager
import java.lang.reflect.Method
import kotlin.reflect.KClass


/**
 *
 * @author NorthLan at 2018/1/29
 */
@Suppress("unused")
@Aspect
@Order(-1) // 保证先于事务执行
@Component
class DataSourceSwitchAspect {
    companion object {
        private val logger = LoggerFactory.getLogger(DataSourceSwitchAspect::class.java)
    }

    init {
        logger.debug("${this::class.java.simpleName} initialized...")
    }

    private var isAnnotationAspect = false

    @Pointcut("execution(* com.gxzc.zen.api..*Service.*(..))")
    fun zenServicePointCut() {
    }

    @Pointcut("execution(* com.gxzc.zen.api..*Mapper.*(..))")
    fun zenMapperPointCut() {
    }

    @Pointcut("execution(* com.baomidou.mybatisplus..*Service.*(..))")
    fun mpServicePointCut() {
    }

    @Pointcut("execution(* com.baomidou.mybatisplus..*Mapper.*(..))")
    fun mpMapperPointCut() {
    }

    /**
     * 一致性事务切面解决方案
     */
    @Around("@annotation(com.gxzc.zen.orm.annotation.MultiTransactional)")
    fun multiTransactionAround(joinPoint: ProceedingJoinPoint): Any? {
        logger.info("@MultiTransactional aspect...")
        val methodName = joinPoint.signature.name
        val parameterTypes = (joinPoint.signature as MethodSignature).method.parameterTypes
        val method = joinPoint.target::class.java.getMethod(methodName, *parameterTypes)

        if (method.isAnnotationPresent(MultiTransactional::class.java)) {
            val multiTransaction = method.getAnnotation(MultiTransactional::class.java)
            val transactionManager = SpringContextHolder.getBean(JtaTransactionManager::class.java)
            val userTransaction = transactionManager?.userTransaction
            return try {
                userTransaction?.begin()
                val r = joinPoint.proceed()
                userTransaction?.commit()
                return r
            } catch (e: Throwable) {
                val rollbackExceptions = multiTransaction.rollbackFor
                val noRollbackExceptions = multiTransaction.noRollbackFor

                val isRollbackExceptionPresent = isPresent(e, rollbackExceptions)
                val isNoRollbackExceptionPresent = isPresent(e, noRollbackExceptions)

                // rollbackFor回滚 noRollbackFor不回滚
                if (isRollbackExceptionPresent || !isNoRollbackExceptionPresent) {
                    userTransaction?.rollback()
                } else {
                    userTransaction?.commit()
                }
            }
        } else {
            return joinPoint.proceed()
        }
    }

    /**
     * 判断指定类型是否存在于注解中
     * @param e 指定抛出类型
     * @param exceptions 注解定义的类型列表
     */
    private fun isPresent(e: Throwable, exceptions: Array<KClass<out Throwable>>): Boolean {
        return exceptions.find {
            e::class.java.isAssignableFrom(it.java) || e::class.java == it::java
        } != null
    }

    @Around("@annotation(com.gxzc.zen.orm.annotation.DynamicDataSource)")
    fun annotationAround(joinPoint: ProceedingJoinPoint): Any? {
        logger.debug("@DynamicDatasource aspect...")
        isAnnotationAspect = true
        val method = (joinPoint.signature as MethodSignature).method
        setDataSource(method, DSKey.DSKEY_SYS)
        try {
            return joinPoint.proceed()
        } finally {
            isAnnotationAspect = false
            DynamicMultipleDataSource.clear()
        }
    }

    @Before("zenMapperPointCut() || mpMapperPointCut()")
    fun dynamicDataSource(joinPoint: JoinPoint) {
        logger.debug("*Mapper aspect...")
        if (isAnnotationAspect) {
            return
        }
        val target = joinPoint.target

        val packageName: String = try {
            val h = target::class.java.superclass.getDeclaredField("h")
            h.isAccessible = true
            val mapperProxy = h.get(target)

            val mapperInterface = mapperProxy.javaClass.getDeclaredField("mapperInterface")
            mapperInterface.isAccessible = true

            val clazz = mapperInterface.get(mapperProxy) as Class<*>

            clazz.`package`.name
        } catch (e: Exception) {
            logger.debug("get packageName of MapperProxy error, ignore if nothing happened!")
            logger.debug("set packageName of target...")
            target::class.java.`package`.name
        }

        matchDataSource((joinPoint.signature as MethodSignature).method, packageName)
    }

    @After("zenMapperPointCut() || mpMapperPointCut()")
    fun afterDynamicDataSource() {
        DynamicMultipleDataSource.clear()
    }

    private fun matchDataSource(method: Method, packageName: String) {
        DSKey.values()
                .filter { it.pkgName in packageName }
                .forEach {
                    setDataSource(method, it.dsKey)
                    logger.debug("setDataSource: {}", it.dsKey)
                }
    }

    /**
     * 设置数据源key
     */
    private fun setDataSource(method: Method, defaultKey: String) {
        val dynamicDataSource: DynamicDataSource? = method.getAnnotation(DynamicDataSource::class.java)
        DynamicMultipleDataSource.setDataSource(dynamicDataSource?.value ?: defaultKey)
    }
}